
/*
 * Compile with:
 * $(CC) -O3  -o c_mannwhitney c_mannwhitney.c bignum.c -lpthread
 * where $(CC) is either gcc or icc
 * bignum.c can be omitted if USE_BIGNUM (below) is 0
 * -lpthread can be omitted if PARALLEL (below) is 0
 */

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <unistd.h>
#include <pthread.h>

// Determine which version is to be compiled:
// PARALLEL == 0 means single-taked; PARALLEL == 1 and FORK == 1 means fork
// processes, otherwise spawn threads.  RECURSIVE is self-explanatory.
// USE_BIGNUM makes the code use the xyssl large-integer implementation;
// otherwise the xyssl functions and types are emulated by #defines.
#define PARALLEL	1
#define FORK		0
#define RECURSIVE	0
#define USE_BIGNUM	1

// Maximum size of bignum data, for passing it through socket
#define NUM_MAXSIZE	1024

#define min(a,b)	((a) > (b)? (b) : (a))

#if USE_BIGNUM
#include "xyssl/bignum.h"
#else
typedef unsigned long long mpi;
#define mpi_init(a, b)
#define mpi_lset(a, b)	(*(a)= (b))
#define mpi_add_int(a, b, c)	(*(a) = *(b) + (c), 0)
#define mpi_add_mpi(a, b, c)	(*(a) = *(b) + *(c), 0)
#define mpi_mul_int(a, b, c)	(*(a) = *(b) * (c), 0)
#define mpi_div_int(a, b, c, d)	(*(a) = *(c) / (d), 0)
#define mpi_write_string(a, b, c, d)	(snprintf((c), *(d), "%llu", *(a)))
#endif

// Padding for all allocations (relevant for multithreaded version)
#define PAD		64
// Preallocation for large integer within thread data struct
#define PREALLOCSZ	64

typedef struct {
  mpi	sum;
  int	outof, m, ugiven;
  int	dstart, dlast, dstride;
  int	threadnr, nthreads;
#if PREALLOCSZ
  char	prealloc[PREALLOCSZ];
#endif
#if PAD
  char	pad[PAD];
#endif
}
nlimitchoicethread;


void *parallelloop( int nthreads, void *threaddatatempl, int threaddatalength,
		void (*threaddatainit)(void *data, int nthreads, int threadnr),
		void *(*threadfunc)(void*), void *(*combine)(void **, int) );
void *forkloop( int nprocs, void *procdatatempl, 
		void (*procdatainit)(void *data, int nprocs, int procnr),
		void *(*procfunc)(void*), int procretsize,
		void *(*combine)(void **, int) );
void par_nlimchoi_init(void *data, int nthreads, int threadnr);
void *par_nlimchoi_thread(void *data);
void *par_nlimchoi_combine(void **data, int n);
void lchoose(mpi *result, int outof, int m);
void nlimitchoices(mpi *sum, int outof, int m, int urest);
void nlimitchoices_i(mpi *sum, int outof, int m, int urest);
void print_mpi(mpi *val);


int main(int argc, char **argv)
{
  nlimitchoicethread templ;
  mpi choices, sum, *result;
  int size, ugiven, d;

  if( argc < 3 ) {
    fprintf(stderr, "usage: c_mannwhitney <size of both samples> <maximal U>\n");
    exit(1);
  }
  size= strtol(argv[1], NULL, 0);
  ugiven= strtol(argv[2], NULL, 0);
  mpi_init(&choices, NULL);
  mpi_lset(&choices, 0);
  lchoose(&choices, 2*size, size);
  print_mpi(&choices);
#if PARALLEL
  templ.outof= 2*size;
  templ.m= size;
  templ.ugiven= ugiven;
#if FORK
  print_mpi( (mpi*)forkloop(10, &templ, par_nlimchoi_init, par_nlimchoi_thread,
		NUM_MAXSIZE, par_nlimchoi_combine) );
#else
  print_mpi( (mpi*)parallelloop(10, &templ, sizeof(nlimitchoicethread),
		par_nlimchoi_init, par_nlimchoi_thread, par_nlimchoi_combine) );
#endif
#else
  mpi_init(&sum, NULL);
  mpi_lset(&sum, 0);
#if RECURSIVE
  nlimitchoices(&sum, 2*size, size, ugiven);
#else
  nlimitchoices_i(&sum, 2*size, size, ugiven);
#endif
  print_mpi(&sum);
#endif
  return 0;
}


/*
 * This function is a general multithreaded paralleliser.  It is passed the
 * number of threads, a template and a length of a data struct to allocate for
 * each thread and pointers to three functions.  The first is called before
 * spawning the threads and is supposed to perform thread-specific
 * initialisation of the thread data struct.  The second is the function which
 * performs one thread's task.  The third is called after all threads have
 * terminated and is supposed to combine the partial results and return the
 * total result in a struct it allocates.
 */
void *parallelloop( int nthreads, void *threaddatatempl, int threaddatalength,
		void (*threaddatainit)(void *data, int nthreads, int threadnr),
		void *(*threadfunc)(void*), void *(*combine)(void **, int) )
{
  pthread_t *threads;
  void *threaddata, *result, **threadresults;
  int threadnr, flag;

  if( nthreads <= 0 )
    nthreads= sysconf(_SC_NPROCESSORS_ONLN);
  threaddata= malloc(nthreads * (sizeof(pthread_t) + threaddatalength + sizeof(void*)));
  if( !threaddata ) {
    fprintf(stderr, "parallelloop: Could not allocate thread data storage.\n");
    return NULL;
  }
  for( threadnr= 0; threadnr< nthreads; ++threadnr )
    memcpy((char*)threaddata + threadnr*threaddatalength,
		threaddatatempl, threaddatalength);
  if( threaddatainit )
    for( threadnr= 0; threadnr< nthreads; ++threadnr )
      (*threaddatainit)((char*)threaddata + threadnr*threaddatalength,
			nthreads, threadnr);
  threads= (pthread_t*)((char*)threaddata + nthreads*threaddatalength);
  for( threadnr= 0; threadnr< nthreads; ++threadnr )
    if( pthread_create(threads+threadnr, NULL, threadfunc, (char*)threaddata+threadnr*threaddatalength) ) {
      fprintf(stderr, "parallelloop: Could not start all threads.  Aborting.\n");
      for( --threadnr, flag= 0; threadnr >= 0; --threadnr )
	flag |= pthread_cancel(threads[threadnr]);
      if( flag )
	fprintf(stderr, "Error cancelling already started threads.\n");
      return NULL;
    }
  threadresults= (void**)(threads + nthreads);
  for( threadnr= 0, flag= 0; threadnr< nthreads; ++threadnr )
    flag |= pthread_join(threads[threadnr], threadresults+threadnr);
  if( flag )
    fprintf(stderr, "parallelloop: Error joining working threads.  Continuing.\n");
  result= NULL;
  if( combine )
    result= (*combine)(threadresults, nthreads);
  free(threaddata);
  return result;
}


/*
 * This function is a forked paralleliser analogous to parallelloop().  Its
 * arguments are similar to those of that function, but for some differences
 * due to forked processes having a different memory space.  The sub-processes'
 * data struct is not allocated by forkloop(), but the template itself becomes
 * each process's struct after forking.  It is passed to the initialisation
 * function before the function performing the process's task is run.  Instead
 * of the size of the process's data struct, forkloop() is passed the size of
 * their return data, which is returned by the worker function and passed
 * through a UNIX socket pair back to the original process.
 */
void *forkloop( int nprocs, void *procdatatempl, 
		void (*procdatainit)(void *data, int nprocs, int procnr),
		void *(*procfunc)(void*), int procretsize,
		void *(*combine)(void **, int) )
{
  void *procdata, *result, **procresults;
  int *sockpairs;
  pid_t fork_ret;
  int procnr, sockret;

  if( nprocs <= 0 )
    nprocs= sysconf(_SC_NPROCESSORS_ONLN);
  procdata= malloc(nprocs * (procretsize + 2*sizeof(int) + sizeof(void*)));
  if( !procdata ) {
    fprintf(stderr, "forkloop: Could not allocate parameter data storage.\n");
    return NULL;
  }
  sockpairs= (int*)((char*)procdata + nprocs*procretsize);
  procresults= (void**)((char*)sockpairs + nprocs*2*sizeof(int));
  procresults[0]= procdata;
  for( procnr= 1; procnr< nprocs; ++procnr )
    procresults[procnr]= (char*)procresults[procnr-1] + procretsize;
  for( procnr= 0; procnr< nprocs; ++procnr ) {
    sockret= socketpair(PF_UNIX, SOCK_STREAM, 0, sockpairs+2*procnr);
    if( sockret < 0 ) {
      fprintf(stderr, "forkloop: Could not generate all socket pairs.  Aborting.\n");
      return NULL;
    }
    fork_ret= fork();
    if( !fork_ret )	// child process
      break;
    if( fork_ret < 0 ) {
      fprintf(stderr, "forkloop: Could not perform all forks.  Aborting.\n");
      return NULL;
    }
  }
  if( !fork_ret ) {
    if( procdatainit )
      (*procdatainit)(procdatatempl, nprocs, procnr);
    result= (*procfunc)(procdatatempl);
    if( send(sockpairs[2*procnr], result, procretsize, 0) != procretsize )
      fprintf(stderr, "forkloop: Forked process %d could not send result.\n", procnr);
    exit(0);
  }
  result= NULL;
  for( procnr= 0; procnr< nprocs; ++procnr ) {
    if( recv(sockpairs[2*procnr+1], procresults[procnr], procretsize, 0) != procretsize ) {
      fprintf(stderr, "forkloop: Could not receive result from forked process %d.  Aborting.\n", procnr);
      return NULL;
    }
  }
  if( combine )
    result= (*combine)(procresults, nprocs);
  free(procdata);
  return result;
}


/*
 * Initialise a thread or sub-process data struct
 */
void par_nlimchoi_init(void *data, int nthreads, int threadnr)
{
  nlimitchoicethread *d= (nlimitchoicethread*)data;

  mpi_init(&d->sum, NULL);
#if USE_BIGNUM && PREALLOCSZ
  d->sum.p= (t_int*)d->prealloc;
  d->sum.n= PREALLOCSZ/sizeof(t_int);
  memset(d->prealloc, 0, PREALLOCSZ);
#else
  mpi_lset(&d->sum, 0);
#endif
  d->dstart= threadnr*threadnr*(d->outof-d->m)/nthreads/nthreads;
  if( threadnr == nthreads - 1 )
    d->dlast= d->outof-d->m;
  else
    d->dlast= (threadnr+1)*(threadnr+1)*(d->outof-d->m)/nthreads/nthreads-1;
  d->dstride= 1;
  d->threadnr= threadnr;
  d->nthreads= nthreads;
}


/*
 * Thread or sub-process worker function
 */
void *par_nlimchoi_thread(void *data)
{
  nlimitchoicethread *dat= (nlimitchoicethread*)data;
  char *forkreturn;
  int t, d;

  t= dat->threadnr;
  if( t >= dat->m )
    return &dat->sum;
  if( t == dat->nthreads - 1 )
#if RECURSIVE
    nlimitchoices(&dat->sum, dat->outof-t, dat->m-t, dat->ugiven);
#else
    nlimitchoices_i(&dat->sum, dat->outof-t, dat->m-t, dat->ugiven);
#endif
  else
    for( d= 1; d<= dat->outof-dat->m; ++d )
#if RECURSIVE
      nlimitchoices(&dat->sum, dat->outof-d-t-1, dat->m-t-1, dat->ugiven-(dat->m-t)*d);
#else
      nlimitchoices_i(&dat->sum, dat->outof-d-t-1, dat->m-t-1, dat->ugiven-(dat->m-t)*d);
#endif
#if FORK
  forkreturn= malloc(NUM_MAXSIZE);
  memcpy(forkreturn, &dat->sum, sizeof(mpi));
  memcpy(forkreturn+sizeof(mpi), dat->sum.p, min(NUM_MAXSIZE-sizeof(mpi), dat->sum.n*sizeof(t_int)));
  return forkreturn;
#else
  return &dat->sum;
#endif
}


/*
 * Add up results of sub-tasks
 */
void *par_nlimchoi_combine(void **data, int n)
{
  static mpi result;

  mpi **sums= (mpi **)data;
  int count;

  mpi_init(&result, NULL);
  mpi_lset(&result, 0);
  for( count= 0; count< n; ++count ) {
#if FORK
    sums[count]->p= (t_int*)((char*)sums[count]+sizeof(mpi));
#endif
    if( mpi_add_mpi(&result, &result, sums[count]) ) {
      fprintf(stderr, "par_nlimchoi_combine: Memory allocation error in bignum\n");
      exit(1);
    }
  }
  return &result;
}


/*
 * Compute binomial coefficient giving the total number of possible choices
 */
void lchoose(mpi *result, int outof, int m)
{
  if( m > outof ) {
    mpi_lset(result, 0);
    return;
  }
  if( !m ) {
    mpi_lset(result, 1);
    return;
  }
  lchoose(result, outof-1, m-1);
  if( mpi_mul_int(result, result, outof) || mpi_div_int(result, NULL, result, m) ) {
    fprintf(stderr, "nlimitchoices: Computation error in bignum\n");
    exit(1);
  }
}


/*
 * Recursive implemenation of computing the number of choices up to a given
 * Mann-Whitney U
 */
void nlimitchoices(mpi *sum, int outof, int m, int urest)
{
  int d;

  if( urest < 0 )
    return;
  if( m == 0 ) {
    if( mpi_add_int(sum, sum, 1) ) {
      fprintf(stderr, "nlimitchoices: Memory allocation error in bignum\n");
      exit(1);
    }
    return;
  }
  for( d= 0; d<= outof-m; ++d )
    nlimitchoices(sum, outof-d-1, m-1, urest-m*d);
}


/*
 * Iterative implemenation of computing the number of choices up to a given
 * Mann-Whitney U
 */
void nlimitchoices_i(mpi *sum, int outof, int m, int urest)
{
  int *dbase;
  int dind, d;

  if( outof < m || urest < 0 )
    return;
//  printf("nlimitchoices_i(.., %d, %d, %d)\n", outof, m, urest);
  dbase= malloc(sizeof(int)*m+PAD);
  if( !dbase ) {
    fprintf(stderr, "nlimitchoices_i: Could not allocate space for backtracking.\n");
    return;
  }
  dind= 0;
  dbase[0]= 0;
  while( dind >= 0 )
  {
    // descend
    while( m > 1 ) {
      ++dind;
      dbase[dind]= 0;
      --outof;
      --m;
    }
    if( mpi_add_int(sum, sum, 1) ) {
      fprintf(stderr, "nlimitchoices_i: Memory allocation error in bignum\n");
      exit(1);
    }
    // backtrack and make different choice
    while( dind >= 0 ) {
      ++dbase[dind];
      --outof;
      urest -= m;
      if( outof >= m && urest >= 0 )
	break;
      outof += dbase[dind]+1;
      urest += m*dbase[dind];
      ++m;
      --dind;
    }
  }
  free(dbase);
//  printf("nlimitchoices_i(.., %d, %d, %d) returns ", outof-1, m-1, urest);
//  print_mpi(sum);
}


/*
 * Print a large integer (xyssl)
 */
#define OUTBUFSZ	1024

void print_mpi(mpi *val)
{
  char outputbuf[OUTBUFSZ];
  int bufsize= OUTBUFSZ;

  mpi_write_string(val, 10, outputbuf, &bufsize);
  printf("%s\n", outputbuf);
}

