
/*
 * Compile with:
 * $(CC) -O3  -o c_mannwhitney_gmp c_mannwhitney_gmp.c -lpthread -lgmp
 * where $(CC) is either gcc or icc
 * -lpthread can be omitted if PARALLEL (below) is 0
 * -lgmp can be omitted if USE_BIGNUM (below) is 0
 */

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <unistd.h>
#include <pthread.h>

// Determine which version is to be compiled:
// PARALLEL == 0 means single-tasked; PARALLEL == 1 and FORK == 1 means fork
// processes, otherwise spawn threads.  RECURSIVE is self-explanatory.
// USE_BIGNUM makes the code use the GMP large-integer implementation;
// otherwise the GMP functions and types are emulated by #defines.
#define PARALLEL	0
#define FORK		0
#define RECURSIVE	0
#define USE_BIGNUM	1

// Maximum size of bignum data, for passing it through socket
#define NUM_MAXSIZE	1024

#define min(a,b)	((a) > (b)? (b) : (a))

#if USE_BIGNUM
#include <gmp.h>
#else
typedef unsigned long long *mpz_t;
#define mpz_init(a)		((a)= (mpz_t)malloc(sizeof(unsigned long long)), *a=0)
#define mpz_set_ui(a, b)	(*(a)= (b))
#define mpz_add_ui(a, b, c)	(*(a) = *(b) + (c))
#define mpz_add(a, b, c)	(*(a) = *(b) + *(c))
#define mpz_mul_ui(a, b, c)	(*(a) = *(b) * (c))
#define mpz_cdiv_q_ui(a, b, c)	(*(a) = *(b) / (c))
#endif

// Preallocation for large integer (not in thread data struct)
#define PREALLOCSZ	64
// Padding for all allocations (relevant for multithreaded version)
#define PAD		64

typedef struct {
  mpz_t	sum;
  int	outof, m, ugiven;
  int	dstart, dlast, dstride;
  int	threadnr, nthreads;
#if PAD
  char	pad[PAD];
#endif
}
nlimitchoicethread;

void *parallelloop( int nthreads, void *threaddatatempl, int threaddatalength,
		void (*threaddatainit)(void *data, int nthreads, int threadnr),
		void *(*threadfunc)(void*), void *(*combine)(void **, int) );
void *forkloop( int nprocs, void *procdatatempl, 
		void (*procdatainit)(void *data, int nprocs, int procnr),
		void *(*procfunc)(void*), int procretsize,
		void *(*combine)(void **, int) );
void par_nlimchoi_init_gmp(void *data, int nthreads, int threadnr);
void *par_nlimchoi_thread_gmp(void *data);
void *par_nlimchoi_combine_gmp(void **data, int n);
void lchoose_gmp(mpz_t result, int outof, int m);
void nlimitchoices_gmp(mpz_t sum, int outof, int m, int urest);
void nlimitchoices_l_gmp(mpz_t sum, int outof, int m, int urest);
void print_mpz(mpz_t val);

int main(int argc, char **argv)
{
  nlimitchoicethread templ;
  mpz_t choices, sum, *result;
  int size, ugiven, d;

  if( argc < 3 ) {
    fprintf(stderr, "usage: c_mannwhitney_gmp <size of both samples> <maximal U>\n");
    exit(1);
  }
  size= strtol(argv[1], NULL, 0);
  ugiven= strtol(argv[2], NULL, 0);
  mpz_init(choices);
  lchoose_gmp(choices, 2*size, size);
  print_mpz(choices);
#if PARALLEL
  templ.outof= 2*size;
  templ.m= size;
  templ.ugiven= ugiven;
#if FORK
  print_mpz( *(mpz_t*)forkloop(10, &templ, par_nlimchoi_init_gmp,
	    par_nlimchoi_thread_gmp, NUM_MAXSIZE, par_nlimchoi_combine_gmp) );
#else
  print_mpz( *(mpz_t*)parallelloop(10, &templ, sizeof(nlimitchoicethread),
   par_nlimchoi_init_gmp, par_nlimchoi_thread_gmp, par_nlimchoi_combine_gmp) );
#endif
#else
  mpz_init(sum);
#if RECURSIVE
  nlimitchoices_gmp(sum, 2*size, size, ugiven);
#else
  nlimitchoices_l_gmp(sum, 2*size, size, ugiven);
#endif
  print_mpz(sum);
#endif
  return 0;
}


/*
 * This function is a general multithreaded paralleliser.  It is passed the
 * number of threads, a template and a length of a data struct to allocate for
 * each thread and pointers to three functions.  The first is called before
 * spawning the threads and is supposed to perform thread-specific
 * initialisation of the thread data struct.  The second is the function which
 * performs one thread's task.  The third is called after all threads have
 * terminated and is supposed to combine the partial results and return the
 * total result in a struct it allocates.
 */
void *parallelloop( int nthreads, void *threaddatatempl, int threaddatalength,
		void (*threaddatainit)(void *data, int nthreads, int threadnr),
		void *(*threadfunc)(void*), void *(*combine)(void **, int) )
{
  pthread_t *threads;
  void *threaddata, *result, **threadresults;
  int threadnr, flag;

  if( nthreads <= 0 )
    nthreads= sysconf(_SC_NPROCESSORS_ONLN);
  threaddata= malloc(nthreads * (sizeof(pthread_t) + threaddatalength + sizeof(void*)));
  if( !threaddata ) {
    fprintf(stderr, "parallelloop: Could not allocate thread data storage.\n");
    return NULL;
  }
  for( threadnr= 0; threadnr< nthreads; ++threadnr )
    memcpy((char*)threaddata + threadnr*threaddatalength,
		threaddatatempl, threaddatalength);
  if( threaddatainit )
    for( threadnr= 0; threadnr< nthreads; ++threadnr )
      (*threaddatainit)((char*)threaddata + threadnr*threaddatalength,
			nthreads, threadnr);
  threads= (pthread_t*)((char*)threaddata + nthreads*threaddatalength);
  for( threadnr= 0; threadnr< nthreads; ++threadnr )
    if( pthread_create(threads+threadnr, NULL, threadfunc, (char*)threaddata+threadnr*threaddatalength) ) {
      fprintf(stderr, "parallelloop: Could not start all threads.  Aborting.\n");
      for( --threadnr, flag= 0; threadnr >= 0; --threadnr )
	flag |= pthread_cancel(threads[threadnr]);
      if( flag )
	fprintf(stderr, "Error cancelling already started threads.\n");
      return NULL;
    }
  threadresults= (void**)(threads + nthreads);
  for( threadnr= 0, flag= 0; threadnr< nthreads; ++threadnr )
    flag |= pthread_join(threads[threadnr], threadresults+threadnr);
  if( flag )
    fprintf(stderr, "parallelloop: Error joining working threads.  Continuing.\n");
  result= NULL;
  if( combine )
    result= (*combine)(threadresults, nthreads);
  free(threaddata);
  return result;
}


/*
 * This function is a forked paralleliser analogous to parallelloop().  Its
 * arguments are similar to those of that function, but for some differences
 * due to forked processes having a different memory space.  The sub-processes'
 * data struct is not allocated by forkloop(), but the template itself becomes
 * each process's struct after forking.  It is passed to the initialisation
 * function before the function performing the process's task is run.  Instead
 * of the size of the process's data struct, forkloop() is passed the size of
 * their return data, which is returned by the worker function and passed
 * through a UNIX socket pair back to the original process.
 */
void *forkloop( int nprocs, void *procdatatempl, 
		void (*procdatainit)(void *data, int nprocs, int procnr),
		void *(*procfunc)(void*), int procretsize,
		void *(*combine)(void **, int) )
{
  void *procdata, *result, **procresults;
  int *sockpairs;
  pid_t fork_ret;
  int procnr, sockret;

  if( nprocs <= 0 )
    nprocs= sysconf(_SC_NPROCESSORS_ONLN);
  procdata= malloc(nprocs * (procretsize + 2*sizeof(int) + sizeof(void*)));
  if( !procdata ) {
    fprintf(stderr, "forkloop: Could not allocate parameter data storage.\n");
    return NULL;
  }
  sockpairs= (int*)((char*)procdata + nprocs*procretsize);
  procresults= (void**)((char*)sockpairs + nprocs*2*sizeof(int));
  procresults[0]= procdata;
  for( procnr= 1; procnr< nprocs; ++procnr )
    procresults[procnr]= (char*)procresults[procnr-1] + procretsize;
  for( procnr= 0; procnr< nprocs; ++procnr ) {
    sockret= socketpair(PF_UNIX, SOCK_STREAM, 0, sockpairs+2*procnr);
    if( sockret < 0 ) {
      fprintf(stderr, "forkloop: Could not generate all socket pairs.  Aborting.\n");
      return NULL;
    }
    fork_ret= fork();
    if( !fork_ret )	// child process
      break;
    if( fork_ret < 0 ) {
      fprintf(stderr, "forkloop: Could not perform all forks.  Aborting.\n");
      return NULL;
    }
  }
  if( !fork_ret ) {
    if( procdatainit )
      (*procdatainit)(procdatatempl, nprocs, procnr);
    result= (*procfunc)(procdatatempl);
    if( send(sockpairs[2*procnr], result, procretsize, 0) != procretsize )
      fprintf(stderr, "forkloop: Forked process %d could not send result.\n", procnr);
    exit(0);
  }
  result= NULL;
  for( procnr= 0; procnr< nprocs; ++procnr ) {
    if( recv(sockpairs[2*procnr+1], procresults[procnr], procretsize, 0) != procretsize ) {
      fprintf(stderr, "forkloop: Could not receive result from forked process %d.  Aborting.\n", procnr);
      return NULL;
    }
  }
  if( combine )
    result= (*combine)(procresults, nprocs);
  free(procdata);
  return result;
}


/*
 * Initialise a thread or sub-process data struct
 */
void par_nlimchoi_init_gmp(void *data, int nthreads, int threadnr)
{
  nlimitchoicethread *d= (nlimitchoicethread*)data;

  mpz_init(d->sum);
#if USE_BIGNUM && PREALLOCSZ
  mpz_realloc2(d->sum, PREALLOCSZ*8);
#endif
  d->dstart= threadnr*threadnr*(d->outof-d->m)/nthreads/nthreads;
  if( threadnr == nthreads - 1 )
    d->dlast= d->outof-d->m;
  else
    d->dlast= (threadnr+1)*(threadnr+1)*(d->outof-d->m)/nthreads/nthreads-1;
  d->dstride= 1;
  d->threadnr= threadnr;
  d->nthreads= nthreads;
}


/*
 * Thread or sub-process worker function
 */
void *par_nlimchoi_thread_gmp(void *data)
{
  nlimitchoicethread *dat= (nlimitchoicethread*)data;
  char *forkreturn;
  size_t exportedsize;
  int t, d;

  t= dat->threadnr;
  if( t >= dat->m )
    return &dat->sum;
  if( t == dat->nthreads - 1 )
#if RECURSIVE
    nlimitchoices_gmp(dat->sum, dat->outof-t, dat->m-t, dat->ugiven);
#else
    nlimitchoices_l_gmp(dat->sum, dat->outof-t, dat->m-t, dat->ugiven);
#endif
  else
    for( d= 1; d<= dat->outof-dat->m; ++d )
#if RECURSIVE
      nlimitchoices_gmp(dat->sum, dat->outof-d-t-1, dat->m-t-1, dat->ugiven-(dat->m-t)*d);
#else
      nlimitchoices_l_gmp(dat->sum, dat->outof-d-t-1, dat->m-t-1, dat->ugiven-(dat->m-t)*d);
#endif
#if FORK
  forkreturn= malloc(NUM_MAXSIZE);
  mpz_export(forkreturn+sizeof(size_t), &exportedsize, 1, 1, 1, 0, dat->sum);
  memcpy(forkreturn, &exportedsize, sizeof(size_t));
  return forkreturn;
#else
  return &dat->sum;
#endif
}


/*
 * Add up results of sub-tasks
 */
void *par_nlimchoi_combine_gmp(void **data, int n)
{
  static mpz_t result;

#if FORK
  char **sums= (char **)data;
  mpz_t sum;
  size_t exportedsize;
#else
  mpz_t **sums= (mpz_t **)data;
#endif
  int count;

  mpz_init(result);
  for( count= 0; count< n; ++count ) {
#if FORK
    exportedsize= *(size_t*)sums[count];
    mpz_import(sum, exportedsize, 1, 1, 1, 0, sums[count]+sizeof(size_t));
    mpz_add(result, result, sum);
#else
    mpz_add(result, result, *sums[count]);
#endif
  }
  return &result;
}


/*
 * Compute binomial coefficient giving the total number of possible choices
 */
void lchoose_gmp(mpz_t result, int outof, int m)
{
  if( m > outof ) {
    mpz_set_ui(result, 0);
    return;
  }
  if( !m ) {
    mpz_set_ui(result, 1);
    return;
  }
  lchoose_gmp(result, outof-1, m-1);
  mpz_mul_ui(result, result, outof);
  mpz_cdiv_q_ui(result, result, m);
}


/*
 * Recursive implemenation of computing the number of choices up to a given
 * Mann-Whitney U
 */
void nlimitchoices_gmp(mpz_t sum, int outof, int m, int urest)
{
  int d;

  if( urest < 0 )
    return;
  if( m == 0 ) {
    mpz_add_ui(sum, sum, 1);
    return;
  }
  for( d= 0; d<= outof-m; ++d )
    nlimitchoices_gmp(sum, outof-d-1, m-1, urest-m*d);
}


/*
 * Iterative implemenation of computing the number of choices up to a given
 * Mann-Whitney U
 */
void nlimitchoices_l_gmp(mpz_t sum, int outof, int m, int urest)
{
  int *dbase;
  int dind, d;

  if( outof < m || urest < 0 )
    return;
//  printf("nlimitchoices_l_gmp(.., %d, %d, %d)\n", outof, m, urest);
  dbase= malloc(sizeof(int)*m+PAD);
  if( !dbase ) {
    fprintf(stderr, "nlimitchoices_l_gmp: Could not allocate space for backtracking.\n");
    return;
  }
  dind= 0;
  dbase[0]= 0;
  while( dind >= 0 )
  {
    // descend
    while( m > 1 ) {
      ++dind;
      dbase[dind]= 0;
      --outof;
      --m;
    }
    mpz_add_ui(sum, sum, 1);
    // backtrack and make different choice
    while( dind >= 0 ) {
      ++dbase[dind];
      --outof;
      urest -= m;
      if( outof >= m && urest >= 0 )
	break;
      outof += dbase[dind]+1;
      urest += m*dbase[dind];
      ++m;
      --dind;
    }
  }
  free(dbase);
//  printf("nlimitchoices_l_gmp(.., %d, %d, %d) returns ", outof-1, m-1, urest);
//  print_mpz(sum);
}


/*
 * Print a large integer (GMP)
 */
void print_mpz(mpz_t val)
{
#if USE_BIGNUM
  char *outputbuf;

  outputbuf= mpz_get_str(NULL, 10, val);
  printf("%s\n", outputbuf);
  free(outputbuf);
#else
  printf("%llu\n", val);
#endif
}

