
/*
 * Compile with:
 * $(CC) -O3  -o c_mannwhitney_memgmp c_mannwhitney_memgmp.c -lgmp
 * where $(CC) is either gcc or icc
 */

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <unistd.h>
#include <pthread.h>

#include <gmp.h>

typedef struct {
  void **cache;
  int *sizes;
  void *oor;
}
memcache;

void lchoose_gmp(mpz_t result, int outof, int m);
void create_cache(memcache *ca, int n, int m, int u);
mpz_t *nlimitchoices_memgmp_mem(memcache *ca, int outof, int m, int urest);
void delete_cache(memcache *ca);
mpz_t *nlimitchoices_memgmp_calc(memcache *ca, int outof, int m, int urest);
void print_mpz(mpz_t val);

int main(int argc, char **argv)
{
  memcache memc;
  mpz_t choices, *result;
  int size, ugiven, d;

  if( argc < 3 ) {
    fprintf(stderr, "usage: countMW <size of both samples> <maximal U>\n");
    exit(1);
  }
  size= strtol(argv[1], NULL, 0);
  ugiven= strtol(argv[2], NULL, 0);
  mpz_init(choices);
  lchoose_gmp(choices, 2*size, size);
  print_mpz(choices);
  create_cache(&memc, 2*size, size, ugiven);
  result= nlimitchoices_memgmp_mem(&memc, 2*size, size, ugiven);
  print_mpz(*result);
  delete_cache(&memc);
  return 0;
}


/*
 * Compute binomial coefficient giving the total number of possible choices
 */
void lchoose_gmp(mpz_t result, int outof, int m)
{
  if( m > outof ) {
    mpz_set_ui(result, 0);
    return;
  }
  if( !m ) {
    mpz_set_ui(result, 1);
    return;
  }
  lchoose_gmp(result, outof-1, m-1);
  mpz_mul_ui(result, result, outof);
  mpz_cdiv_q_ui(result, result, m);
}


/*
 * Set up cache for values of nlimitchoices() up to the given n, m and u
 */
void create_cache(memcache *ca, int n, int m, int u)
{
  void **clear;

  ca->sizes= (int *)malloc(3*sizeof(int));
  ca->sizes[0]= n+1;
  ca->sizes[1]= ca->sizes[0]*(m+1);
  ca->sizes[2]= ca->sizes[1]*(u+1);
  ca->cache= malloc(ca->sizes[2]*sizeof(mpz_t*));
  for( clear= ca->cache; clear< ca->cache+ca->sizes[2]; ++clear )
    *clear= NULL;
  ca->oor= malloc(sizeof(mpz_t));
  mpz_init(*(mpz_t*)ca->oor);
}


/*
 * Try to find value of nlimitchoices() in the cache, otherwise call the
 * calculation function nlimitchoices_memgmp_calc().
 */
mpz_t *nlimitchoices_memgmp_mem(memcache *ca, int outof, int m, int urest)
{
  int index;

  if( urest < 0 )
    return (mpz_t*)ca->oor;
  index= outof+ca->sizes[0]*m+ca->sizes[1]*urest;
  if( !ca->cache[index] )
    ca->cache[index]= nlimitchoices_memgmp_calc(ca, outof, m, urest);
  return (mpz_t*)ca->cache[index];
}


/*
 * Delete the cache
 */
void delete_cache(memcache *ca)
{
  void **cached;

  for( cached= ca->cache; cached< ca->cache+ca->sizes[2]; ++cached )
    if( *cached )
      free(*cached);
  free(ca->cache);
  free(ca->sizes);
}


/*
 * Compute value of nlimitchoices() for the given arguments, but call the
 * cache lookup function nlimitchoices_memgmp_mem() when recursing.
 */
mpz_t *nlimitchoices_memgmp_calc(memcache *ca, int outof, int m, int urest)
{
  mpz_t *res= (mpz_t*)malloc(sizeof(mpz_t));
  int d;

  mpz_init(*res);
  if( urest < 0 )
    return res;
  if( m == 0 ) {
    mpz_set_ui(*res, 1);
    return res;
  }
  for( d= 0; d<= outof-m; ++d ) {
    mpz_t *sub= nlimitchoices_memgmp_mem(ca, outof-d-1, m-1, urest-m*d);
    mpz_add(*res, *res, *sub);
  }
  return res;
}


/*
 * Print a large integer (GMP)
 */
void print_mpz(mpz_t val)
{
  char *outputbuf;

  outputbuf= mpz_get_str(NULL, 10, val);
  printf("%s\n", outputbuf);
  free(outputbuf);
}

