
/*
 * Compile with:
 * $(CC) -O3  -o c_mannwhitney c_mannwhitney.c bignum.c
 * where $(CC) is either gcc or icc
 */

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <unistd.h>

// Determine which version is to be compiled:
// RECURSIVE is self-explanatory.
// USE_BIGNUM makes the code use the xyssl large-integer implementation;
// otherwise the xyssl functions and types are emulated by #defines.
#define RECURSIVE	0
#define USE_BIGNUM	1

#define min(a,b)	((a) > (b)? (b) : (a))

#if USE_BIGNUM
#include "xyssl/bignum.h"
#else
typedef unsigned long long mpi;
#define mpi_init(a, b)
#define mpi_lset(a, b)	(*(a)= (b))
#define mpi_add_int(a, b, c)	(*(a) = *(b) + (c), 0)
#define mpi_add_mpi(a, b, c)	(*(a) = *(b) + *(c), 0)
#define mpi_mul_int(a, b, c)	(*(a) = *(b) * (c), 0)
#define mpi_div_int(a, b, c, d)	(*(a) = *(c) / (d), 0)
#define mpi_write_string(a, b, c, d)	(snprintf((c), *(d), "%llu", *(a)))
#endif

void lchoose(mpi *result, int outof, int m);
void nlimitchoices(mpi *sum, int outof, int m, int urest);
void nlimitchoices_i(mpi *sum, int outof, int m, int urest);
void print_mpi(mpi *val);

int main(int argc, char **argv)
{
  mpi choices, sum, *result;
  int size, n, m, ugiven, level, t, d;

  if( argc < 4 ) {
    fprintf(stderr, "usage: c_mannwhitney_part <size of both samples> <maximal U> <level>\n"
		"level < 0 if last (ie loop over d including 0).\n");
    exit(1);
  }
  size= strtol(argv[1], NULL, 0);
  ugiven= strtol(argv[2], NULL, 0);
  level= strtol(argv[3], NULL, 0);
  n= 2*size;
  m= size;
  t= abs(level);
  if( level == 0 ) {
    mpi_init(&choices, NULL);
    mpi_lset(&choices, 0);
    lchoose(&choices, n, m);
    print_mpi(&choices);
  }
  mpi_init(&sum, NULL);
  mpi_lset(&sum, 0);
  if( level < 0 )
#if RECURSIVE
    nlimitchoices(&sum, n-t, m-t, ugiven);
#else
    nlimitchoices_i(&sum, n-t, m-t, ugiven);
#endif
  else
    for( d= 1; d<= n-m; ++d )
#if RECURSIVE
      nlimitchoices(&sum, n-d-t-1, m-t-1, ugiven-(m-t)*d);
#else
      nlimitchoices_i(&sum, n-d-t-1, m-t-1, ugiven-(m-t)*d);
#endif
  print_mpi(&sum);
  return 0;
}



/*
 * Compute binomial coefficient giving the total number of possible choices
 */
void lchoose(mpi *result, int outof, int m)
{
  if( m > outof ) {
    mpi_lset(result, 0);
    return;
  }
  if( !m ) {
    mpi_lset(result, 1);
    return;
  }
  lchoose(result, outof-1, m-1);
  if( mpi_mul_int(result, result, outof) || mpi_div_int(result, NULL, result, m) ) {
    fprintf(stderr, "nlimitchoices: Computation error in bignum\n");
    exit(1);
  }
}


/*
 * Recursive implemenation of computing the number of choices up to a given
 * Mann-Whitney U
 */
void nlimitchoices(mpi *sum, int outof, int m, int urest)
{
  int d;

  if( urest < 0 )
    return;
  if( m == 0 ) {
    if( mpi_add_int(sum, sum, 1) ) {
      fprintf(stderr, "nlimitchoices: Memory allocation error in bignum\n");
      exit(1);
    }
    return;
  }
  for( d= 0; d<= outof-m; ++d )
    nlimitchoices(sum, outof-d-1, m-1, urest-m*d);
}


/*
 * Iterative implemenation of computing the number of choices up to a given
 * Mann-Whitney U
 */
void nlimitchoices_i(mpi *sum, int outof, int m, int urest)
{
  int *dbase;
  int dind, d;

  if( outof < m || urest < 0 )
    return;
//  printf("nlimitchoices_i(.., %d, %d, %d)\n", outof, m, urest);
  dbase= malloc(sizeof(int)*m);
  if( !dbase ) {
    fprintf(stderr, "nlimitchoices_i: Could not allocate space for backtracking.\n");
    return;
  }
  dind= 0;
  dbase[0]= 0;
  while( dind >= 0 )
  {
    // descend
    while( m > 1 ) {
      ++dind;
      dbase[dind]= 0;
      --outof;
      --m;
    }
    if( mpi_add_int(sum, sum, 1) ) {
      fprintf(stderr, "nlimitchoices_i: Memory allocation error in bignum\n");
      exit(1);
    }
    // backtrack and make different choice
    while( dind >= 0 ) {
      ++dbase[dind];
      --outof;
      urest -= m;
      if( outof >= m && urest >= 0 )
	break;
      outof += dbase[dind]+1;
      urest += m*dbase[dind];
      ++m;
      --dind;
    }
  }
  free(dbase);
//  printf("nlimitchoices_i(.., %d, %d, %d) returns ", outof-1, m-1, urest);
//  print_mpi(sum);
}


/*
 * Print a large integer (xyssl)
 */
#define OUTBUFSZ	1024

void print_mpi(mpi *val)
{
  char outputbuf[OUTBUFSZ];
  int bufsize= OUTBUFSZ;

  mpi_write_string(val, 10, outputbuf, &bufsize);
  printf("%s\n", outputbuf);
}

